@@ -1,6 +1,6 @@
 /*
 buildtorrent -- torrent file creation program
-Copyright (C) 2007,2008 Claude Heiland-Allen
+Copyright (C) 2007,2008,2009,2010 Claude Heiland-Allen
 
 
 This program is free software; you can redistribute it and/or
@@ -38,7 +38,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 /******************************************************************************
 program version string
 ******************************************************************************/
-const char* __version__ = "0.7";
+#define bt_version "0.8"
+const char* __version__ = bt_version;
 
 /******************************************************************************
 torrent data structure declarations
@@ -55,7 +56,8 @@ enum _bt_error {
   BT_ERROR_TYPE,
   BT_ERROR_MEMORY,
   BT_ERROR_IO,
-  BT_ERROR_OVERFLOW
+  BT_ERROR_OVERFLOW,
+  BT_ERROR_PARSE
 };
 typedef enum _bt_error bt_error;
 
@@ -636,46 +638,46 @@ void bt_show(bt_data bd, int piecestoo, int indent, int indentstep, int comma) {
 
 
 /******************************************************************************
-hash queue data structure
+file list data structure
 the first in the list is a dummy
 ******************************************************************************/
-struct _bt_hash_queue {
-  integer size;
+struct _bt_file_list {
   char *file;
-  struct _bt_hash_queue *next;
+  bt_data path;
+  struct _bt_file_list *next;
 };
-typedef struct _bt_hash_queue *bt_hash_queue;
+typedef struct _bt_file_list *bt_file_list;
 
 /******************************************************************************
 create the dummy first list element
 ******************************************************************************/
-bt_hash_queue bt_create_hash_queue() {
-  bt_hash_queue queue = malloc(sizeof(struct _bt_hash_queue));
-  if (!queue) {
+bt_file_list bt_create_file_list() {
+  bt_file_list flist = malloc(sizeof(struct _bt_file_list));
+  if (!flist) {
     return NULL;
   }
-  queue->size = 0;
-  queue->file = NULL;
-  queue->next = NULL;
-  return queue;
+  flist->file = NULL;
+  flist->path = NULL;
+  flist->next = NULL;
+  return flist;
 }
 
 /******************************************************************************
-append to the hash queue
+prepend to the file list
 copies the arguments
+path may be NULL, but only in single file mode
 ******************************************************************************/
-bt_error bt_hash_enqueue(
-  bt_hash_queue queue, const char *file, integer size, int verbose
+bt_error bt_file_list_prepend(
+  bt_file_list flist, const char *file, bt_data path
 ) {
-  if (!queue) {
+  if (!flist) {
     return BT_ERROR_NULL;
   }
   if (!file) {
     return BT_ERROR_NULL;
   }
   {
-  bt_hash_queue current;
-  bt_hash_queue node = malloc(sizeof(struct _bt_hash_queue));
+  bt_file_list node = malloc(sizeof(struct _bt_file_list));
   if (!node) {
     return BT_ERROR_MEMORY;
   }
@@ -684,20 +686,84 @@ bt_error bt_hash_enqueue(
     return BT_ERROR_MEMORY;
   }
   memcpy(node->file, file, strlen(file) + 1);
-  node->size = size;
-  node->next = NULL;
-  current = queue;
-  while (current->next) {
-    current = current->next;
-  }
-  current->next = node;
-  if (verbose) {
-    fprintf(stderr, "%24" PRId64 " : %s\n", size, file);
+  if (path) {
+    if (!(node->path = bt_copy(path))) {
+      free(node->file);
+      free(node);
+      return BT_ERROR_MEMORY;
+    }
+  } else {
+    node->path = NULL;
   }
+  node->next = flist->next;
+  flist->next = node;
   return BT_ERROR_NONE;
   }
 }
 
+
+/******************************************************************************
+annotated file list data structure
+the first in the list is a dummy
+******************************************************************************/
+struct _bt_afile_list {
+  char *file;
+  bt_data path;
+  integer length;
+  bt_data md5sum;
+  struct _bt_afile_list *next;
+};
+typedef struct _bt_afile_list *bt_afile_list;
+
+/******************************************************************************
+create the dummy first list element
+******************************************************************************/
+bt_afile_list bt_create_afile_list() {
+  bt_afile_list aflist = malloc(sizeof(struct _bt_afile_list));
+  if (!aflist) {
+    return NULL;
+  }
+  aflist->file = NULL;
+  aflist->path = NULL;
+  aflist->length = 0;
+  aflist->md5sum = NULL;
+  aflist->next = NULL;
+  return aflist;
+}
+
+/******************************************************************************
+prepend to the annotated file list
+aflist is an annotated file list to prepend to
+fnode is a file list node to annotate
+length is the length to annotate with
+md5sum is the md5sum to annotate with (may be NULL)
+NOTE does NOT create a new copy of the existing data
+******************************************************************************/
+bt_error bt_afile_list_prepend(
+  bt_afile_list aflist, bt_file_list fnode, integer length, bt_data md5sum
+) {
+  if (!aflist) {
+    return BT_ERROR_NULL;
+  }
+  if (!fnode) {
+    return BT_ERROR_NULL;
+  }
+  {
+  bt_afile_list node = malloc(sizeof(struct _bt_afile_list));
+  if (!node) {
+    return BT_ERROR_MEMORY;
+  }
+  node->file = fnode->file;
+  node->path = fnode->path;
+  node->length = length;
+  node->md5sum = md5sum;
+  node->next = aflist->next;
+  aflist->next = node;
+  return BT_ERROR_NONE;
+  }
+}
+
+
 /******************************************************************************
 append a file to a path string
 path must be at least length maxlength bytes
@@ -722,7 +788,6 @@ bt_error bt_joinpath(size_t maxlength, char *path, const char *file) {
 /******************************************************************************
 md5sum a file
 ******************************************************************************/
-
 bt_data bt_md5sum_file(const char *filename, integer length) {
   struct MD5Context context;
   char *hexdigits = "0123456789abcdef";
@@ -770,63 +835,46 @@ bt_data bt_md5sum_file(const char *filename, integer length) {
 }
 
 
-
 /******************************************************************************
 find files recursively
 directory is the initial directory
 path is a buffer of maxlength bytes to store the accumulated path
 pathlist is the accumulated path exploded as a list
-files is a list to add the found files to
+files is a file list to add the found files to
 returns an error flag
 ******************************************************************************/
-bt_error bt_findfiles(
-  DIR *directory, size_t maxlength, char *path, bt_data pathlist,
-  bt_data files, bt_hash_queue queue, int domd5sum, int verbose
+bt_error bt_find_files(
+  DIR *directory, size_t maxlength, char *path, bt_data pathlist, bt_file_list files
 ) {
   struct dirent *entry;
-  bt_data file = NULL;
   bt_data filename = NULL;
-  bt_data filesize = NULL;
   bt_data filepath = NULL;
-  bt_data md5sum = NULL;
   bt_error err = BT_ERROR_NONE;
   while ((entry = readdir(directory))) {
     if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..")) {
       struct stat s;
       size_t oldlength = strlen(path);
-      if ((err = bt_joinpath(maxlength, path, entry->d_name))) {
-        return err;
-      }
-      if (!stat(path, &s)) {
-        if (S_ISREG(s.st_mode)) {
-          if ((file = bt_dictionary())) {
-            if ((filesize = bt_integer(s.st_size))) {
-              if ((
-                filename = bt_string(strlen(entry->d_name), entry->d_name)
-              )) {
+      if (!(err = bt_joinpath(maxlength, path, entry->d_name))) {
+        if (!stat(path, &s)) {
+          if (S_ISREG(s.st_mode)) {
+            if ((filename = bt_string(strlen(entry->d_name), entry->d_name))) {
+              if ((filepath = bt_copy(pathlist))) {
+                if (!(err = bt_list_append(filepath, filename))) {
+                  err = bt_file_list_prepend(files, path, filepath);
+                }
+              } else {
+                err = BT_ERROR_MEMORY;
+              }
+            } else {
+              err = BT_ERROR_MEMORY;
+            }
+          } else if (S_ISDIR(s.st_mode)) {
+            DIR* dir;
+            if ((dir = opendir(path))) {
+              if ((filename = bt_string(strlen(entry->d_name), entry->d_name))) {
                 if ((filepath = bt_copy(pathlist))) {
                   if (!(err = bt_list_append(filepath, filename))) {
-                    if (!(err = bt_dictionary_insert(
-                      file, strlen("length"), "length", filesize
-                    ))) {
-    if (domd5sum) {
-      if ((md5sum = bt_md5sum_file(path, s.st_size))) {
-        err = bt_dictionary_insert(file, strlen("md5sum"), "md5sum", md5sum);
-      } else {
-        err = BT_ERROR_IO;
-      }
-    }
-    if (!err) {
-    
-                      if (!(err = bt_dictionary_insert(
-                        file, strlen("path"), "path", filepath
-                      ))) {
-                        if (!(err = bt_list_append(files, file))) {
-                          err = bt_hash_enqueue(queue, path, s.st_size, verbose);
-                        }
-                      }
-                    }
-    }
+                    err = bt_find_files(dir, maxlength, path, filepath, files);
                   }
                 } else {
                   err = BT_ERROR_MEMORY;
@@ -834,49 +882,106 @@ bt_error bt_findfiles(
               } else {
                 err = BT_ERROR_MEMORY;
               }
+              closedir(dir);
             } else {
-              err = BT_ERROR_MEMORY;
+              err = BT_ERROR_IO;
             }
           } else {
-            err = BT_ERROR_MEMORY;
-          }
-        } else if (S_ISDIR(s.st_mode)) {
-          DIR* dir;
-          if ((dir = opendir(path))) {
-            if ((
-              filename = bt_string(strlen(entry->d_name), entry->d_name)
-            )) {
-              if ((filepath = bt_copy(pathlist))) {
-                if (!(err = bt_list_append(filepath, filename))) {
-                  err = bt_findfiles(
-                    dir, maxlength, path, filepath, files, queue, domd5sum, verbose
-                  );
-                  closedir(dir);
-                }
-              } else {
-                err = BT_ERROR_MEMORY;
-              }
-            } else {
-              err = BT_ERROR_MEMORY;
-            }
-          } else {
-            err = BT_ERROR_IO;
+            /* FIXME neither regular file nor directory, what to do? */
           }
+        } else {
+          err = BT_ERROR_IO;
         }
-      } else {
-        err = BT_ERROR_IO;
       }
       path[oldlength] = '\0';
     }
+    if (err) {
+      break;
+    }
   }
   return err;
 }
 
 
+/******************************************************************************
+annotate a file list
+files is the file list to annotate
+afiles is the file list to add to
+returns an error flag
+******************************************************************************/
+bt_error bt_annotate_files(
+  bt_file_list files, bt_afile_list afiles
+) {
+  bt_error err = BT_ERROR_NONE;
+  bt_file_list fnode = files;
+  while ((fnode = fnode->next)) {
+    struct stat s;
+    if (!stat(fnode->file, &s)) {
+      if (S_ISREG(s.st_mode)) {
+        integer length = s.st_size;
+        err = bt_afile_list_prepend(afiles, fnode, length, NULL);
+      } else {
+        err = BT_ERROR_IO;
+      }
+    } else {
+      err = BT_ERROR_IO;
+    }
+    if (err) {
+      break;
+    }
+  }
+  return err;
+}
+
+
+/******************************************************************************
+convert an annotated file list into torrent format
+aflist is the list to convert
+returns the torrent format file list
+******************************************************************************/
+bt_data bt_afile_list_info(bt_afile_list aflist) {
+  bt_afile_list node;
+  bt_data files;
+  if (!(files = bt_list())) {
+    return NULL;
+  }
+  node = aflist;
+  while ((node = node->next)) {
+    bt_data file;
+    bt_data filesize;
+    bt_data filepath;
+    bt_data md5sum;
+    if (!(file = bt_dictionary())) {
+      return NULL;
+    }
+    if (!(filesize = bt_integer(node->length))) {
+      return NULL;
+    }
+    if (!(filepath = node->path)) {
+      return NULL;
+    }
+    if (bt_dictionary_insert(file, strlen("length"), "length", filesize)) {
+      return NULL;
+    }
+    if (bt_dictionary_insert(file, strlen("path"), "path", filepath)) {
+      return NULL;
+    }
+    if ((md5sum = node->md5sum)) {
+      if (bt_dictionary_insert(file, strlen("md5sum"), "md5sum", md5sum)) {
+        return NULL;
+      }
+    }
+    if (bt_list_append(files, file)) {
+      return NULL;
+    }
+  }
+  return files;
+}
+
+
 /******************************************************************************
 draw progressbar
 ******************************************************************************/
-
 void bt_progressbar(integer piece, integer count) {
   integer oldblocks = ((piece - 1) * 50) / count;
   integer blocks = (piece * 50) / count;
@@ -905,10 +1010,12 @@ void bt_progressbar(integer piece, integer count) {
 
 /******************************************************************************
 hash files
+return piece hash as a string
+data side effect: if domd5sum then add md5sums to aflist
+output side effect: if verbose then show file summary
 ******************************************************************************/
-
-bt_data bt_hash_pieces(bt_hash_queue queue, integer size, int verbose) {
-  bt_hash_queue node = NULL;
+bt_data bt_hash_pieces(bt_afile_list aflist, integer size, int domd5sum, int verbose) {
+  bt_afile_list node = NULL;
   char *hashdata = NULL;
   integer total = 0;
   unsigned char *buffer = NULL;
@@ -919,16 +1026,23 @@ bt_data bt_hash_pieces(bt_hash_queue queue, integer size, int verbose) {
   integer piececount;
   integer i;
   sha1_byte digest[SHA1_DIGEST_LENGTH];
-  if (!queue) {
+  if (!aflist) {
     return NULL;
   }
   if (!(buffer = malloc(size))) {
     return NULL;
   }
-  node = queue;
-  while (node->next) {
-    node = node->next;
-    total += node->size;
+  node = aflist;
+  while ((node = node->next)) {
+    if (verbose) {
+      fprintf(stderr, "%20" PRId64 " : %s\n", node->length, node->file);
+    }
+    if (domd5sum) {
+      if (!(node->md5sum = bt_md5sum_file(node->file, node->length))) {
+        fprintf(stderr, "buildtorrent: error computing md5sum for \"%s\"\n", node->file);
+      }
+    }
+    total += node->length;
   }
   piececount = (total + size - 1) / size; /* ceil(total/size) */
   if (piececount <= 0) { /* FIXME: no idea what to do if there's no data */
@@ -950,13 +1064,13 @@ bt_data bt_hash_pieces(bt_hash_queue queue, integer size, int verbose) {
     free(buffer);
     return NULL;
   }
-  node = queue->next;
+  node = aflist->next;
   file = fopen(node->file, "rb");
   if (!file) {
     free(buffer);
     return NULL;
   }
-  left = node->size;
+  left = node->length;
   bufptr = buffer;
   for (i = 0; i < piececount; ++i) {
     do {
@@ -980,7 +1094,7 @@ bt_data bt_hash_pieces(bt_hash_queue queue, integer size, int verbose) {
             free(buffer);
             return NULL;
           }
-          left = node->size;
+          left = node->length;
         }
       } else { /* left > remain */
         /* take as much as we can */
@@ -1019,20 +1133,11 @@ bt_data bt_hash_pieces(bt_hash_queue queue, integer size, int verbose) {
   return bt_string(SHA1_DIGEST_LENGTH * piececount, hashdata);
 }
 
-/******************************************************************************
-parse bencoded data, eg a torrent file
-******************************************************************************/
-/*
-bt_data bt_parse_torrent(FILE *in) {
-  
-}
-*/
 
 /******************************************************************************
 parse an announce list
 format = "url|url|url,url|url|url,url|url|url"
 ******************************************************************************/
-
 bt_data bt_parse_announcelist(const char *urls) {
   bt_data announcelist;
   bt_data tier;
@@ -1056,12 +1161,16 @@ bt_data bt_parse_announcelist(const char *urls) {
     t = NULL;
     t1 = strchr(s, '|');
     t2 = strchr(s, ',');
-    if (!(t1 || t2)) {
+    if (!t1 && !t2) {
       t = s + strlen(s);
+    } else if (!t1) {
+      t = t2;
+    } else if (!t2) {
+      t = t1;
     } else {
-      t = (t1 - t2) > 0 ? t2 : t1;
+      t = (t1 < t2) ? t1 : t2;
     }
-    if (t - s == 0) {
+    if (t <= s) {
       return NULL;
     }
     if (bt_list_append(tier, bt_string(t - s, s))) {
@@ -1082,27 +1191,190 @@ bt_data bt_parse_announcelist(const char *urls) {
   return announcelist;
 }
 
+/******************************************************************************
+parse a webseed list
+format = "url,url,url"
+******************************************************************************/
+bt_data bt_parse_webseedlist(const char *urls) {
+  bt_data webseedlist;
+  const char *s;
+  const char *t;
+  const char *t2;
+  if (!urls) {
+    return NULL;
+  }
+  if (strcmp("", urls) == 0) {
+    return NULL;
+  }
+  webseedlist = bt_list();
+  if (!webseedlist) {
+    return NULL;
+  }
+  s = urls;
+  do {
+    t = NULL;
+    t2 = strchr(s, ',');
+    if (!t2) {
+      t = s + strlen(s);
+    } else {
+      t = t2;
+    }
+    if (t <= s) {
+      return NULL;
+    }
+    if (bt_list_append(webseedlist, bt_string(t - s, s))) {
+      return NULL;
+    }
+    s = t + 1;
+  } while (t[0] != '\0');
+  return webseedlist;
+}
+
+/******************************************************************************
+read and parse a filelist file
+line format = "real/file/system/path|tor/rent/path/file"
+delimiters: '|', '\n'
+escape character: '\\'
+******************************************************************************/
+enum _bt_parse_state {
+  BT_PARSE_LEFT = 0,
+  BT_PARSE_LEFT_ESCAPED,
+  BT_PARSE_RIGHT,
+  BT_PARSE_RIGHT_ESCAPED,
+  BT_PARSE_OK,
+  BT_PARSE_ERROR
+};
+typedef enum _bt_parse_state bt_parse_state;
+bt_error bt_parse_filelist(bt_file_list flist, FILE *infile) {
+  char filename[8192];
+  char torrname[8192];
+  bt_data torrpath = NULL;
+  int c;
+  int i = 0;
+  int state = BT_PARSE_LEFT;
+  while(state != BT_PARSE_OK && state != BT_PARSE_ERROR) {
+    c = getc(infile);
+    switch (state) {
+    case (BT_PARSE_LEFT):
+      if (c < 0) {
+        state = BT_PARSE_OK;
+      } else if (c == '\\') {
+        state = BT_PARSE_LEFT_ESCAPED;
+      } else if (c == '|') {
+        filename[i] = '\0';
+        i = 0;
+        torrpath = bt_list();
+        if (!torrpath) {
+          state = BT_PARSE_ERROR;
+        } else {
+          state = BT_PARSE_RIGHT;
+        }
+      } else {
+        filename[i++] = c;
+        if (i > 8190) {
+          state = BT_PARSE_ERROR;
+        } else {
+          state = BT_PARSE_LEFT;
+        }
+      }
+      break;
+    case (BT_PARSE_LEFT_ESCAPED):
+      if (c < 0) {
+        state = BT_PARSE_ERROR;
+      } else {
+        filename[i++] = c;
+        if (i > 8190) {
+          state = BT_PARSE_ERROR;
+        } else {
+          state = BT_PARSE_LEFT;
+        }
+      }
+      break;
+    case (BT_PARSE_RIGHT):
+      if (c < 0) {
+        state = BT_PARSE_ERROR;
+      } else if (c == '\\') {
+        state = BT_PARSE_RIGHT_ESCAPED;
+      } else if (c == '/' || c == '\n') {
+        bt_data torrnamestr;
+        torrname[i] = '\0';
+        i = 0;
+        if (0 == strcmp("", torrname)) {
+          state = BT_PARSE_ERROR;
+        } else {
+          torrnamestr = bt_string(strlen(torrname), torrname);
+          if (!torrnamestr) {
+            state = BT_PARSE_ERROR;
+          } else {
+            bt_list_append(torrpath, torrnamestr);
+            if (c == '\n') {
+              bt_file_list_prepend(flist, filename, torrpath);
+              state = BT_PARSE_LEFT;
+            } else {
+              state = BT_PARSE_RIGHT;
+            }
+          }
+        }
+      } else {
+        torrname[i++] = c;
+        if (i > 8190) {
+          state = BT_PARSE_ERROR;
+        } else {
+          state = BT_PARSE_RIGHT;
+        }
+      }
+      break;
+    case (BT_PARSE_RIGHT_ESCAPED):
+      if (c < 0) {
+        state = BT_PARSE_ERROR;
+      } else {
+        torrname[i++] = c;
+        if (i > 8190) {
+          state = BT_PARSE_ERROR;
+        } else {
+          state = BT_PARSE_RIGHT;
+        }
+      }
+      break;
+    default:
+      fprintf(stderr, "buildtorrent: internal error parsing file list (%d)\n", state);
+      break;
+    }
+  }
+  if (state == BT_PARSE_OK) {
+    return BT_ERROR_NONE;
+  } else {
+    return BT_ERROR_PARSE;
+  }
+}
+
 /******************************************************************************
 show usage message
 ******************************************************************************/
 void bt_usage(void) {
   printf(
-    "Usage: buildtorrent [OPTIONS] -a announceurl input output\n"
+    "Usage:\n"
+    "  buildtorrent [OPTIONS] -a announceurl input output\n"
+    "  buildtorrent [OPTIONS] -a announceurl -f filelist -n name output\n"
     "\n"
     "options:\n"
-    "--announce      -a  <announce>   : announce url (required)\n"
-    "--announcelist  -A  <announces>  : announce url list\n"
-    "--piecelength   -l  <length>     : piece length in bytes, default 262144\n"
-    "--comment       -c  <comment>    : user comment, omitted by default\n"
-    "--private       -p  <private>    : private flag, either 0 or 1\n"
-    "--nodate        -D               : omit 'creation date' field\n"
-    "--nocreator     -C               : omit 'created by' field\n"
-    "--md5sum        -m               : add an 'md5sum' field for each file\n"
-    "--show          -s               : show generated torrent structure\n"
-    "--showall       -S               : show pieces too (implies '-s')\n"
-    "--quiet         -q               : quiet operation\n"
-    "--version       -V               : show version of buildtorrent\n"
-    "--help          -h               : show this help screen\n"
+    "--announce        -a  <announce>   : announce url (required)\n"
+    "--filelist        -f  <filelist>   : external file list (requires '-n')\n"
+    "--name            -n  <name>       : torrent name, default based on input\n"
+    "--announcelist    -A  <announces>  : announce url list (format: a,b1|b2,c)\n"
+    "--webseeds        -w  <webseeds>   : webseed url list (format: a,b,c)\n"
+    "--piecelength     -l  <length>     : piece length in bytes, default 262144\n"
+    "--piecesize       -L  <size>       : use 2^size as piece length, default 18\n"
+    "--comment         -c  <comment>    : user comment, omitted by default\n"
+    "--private         -p  <private>    : private flag, either 0 or 1\n"
+    "--nodate          -D               : omit 'creation date' field\n"
+    "--nocreator       -C               : omit 'created by' field\n"
+    "--md5sum          -m               : add an 'md5sum' field for each file\n"
+    "--show            -s               : show generated torrent structure\n"
+    "--showall         -S               : show pieces too (implies '-s')\n"
+    "--quiet           -q               : quiet operation\n"
+    "--version         -V               : show version of buildtorrent\n"
+    "--help            -h               : show this help screen\n"
   );
 }
 
@@ -1113,10 +1385,14 @@ int main(int argc, char **argv) {
 
   char *url = NULL;
   char *urls = NULL;
+  char *wurls = NULL;
   char *inname = NULL;
+  char *nameflag = NULL;
   char *namebase = NULL;
   char *outfile = NULL;
   char *commentstr = NULL;
+  char *filelistfilename = NULL;
+  int lplen = -1;
   unsigned int plen = 262144;
   int verbose = 1;
   int nodate = 0;
@@ -1128,24 +1404,28 @@ int main(int argc, char **argv) {
   int slen;
   int i;
 
-  DIR *dir;
-  FILE *output;
-  bt_data torrent;
-  bt_data announce;
+  DIR *dir = NULL;
+  FILE *output = NULL;
+  bt_data torrent = NULL;
+  bt_data announce = NULL;
   bt_data announcelist = NULL;
-  bt_data info;
-  bt_data piecelength;
-  bt_data pieces;
-  bt_data files;
-  bt_data name;
-  bt_data length;
-  bt_data pathlist;
-  bt_data creator;
-  bt_data creationdate;
-  bt_data comment;
-  bt_data private;
-  bt_data md5sum;
-  bt_hash_queue queue;
+  bt_data webseedlist = NULL;
+  bt_data info = NULL;
+  bt_data piecelength = NULL;
+  bt_data pieces = NULL;
+  bt_data files = NULL;
+  bt_data name = NULL;
+  bt_data length = NULL;
+  bt_data pathlist = NULL;
+  bt_data creator = NULL;
+  bt_data creationdate = NULL;
+  bt_data comment = NULL;
+  bt_data private = NULL;
+  bt_data md5sum = NULL;
+
+  int multifile = 0;
+  bt_file_list flist = NULL;
+  bt_afile_list aflist = NULL;
 
   struct stat s;
   char path[8192];
@@ -1155,8 +1435,12 @@ int main(int argc, char **argv) {
     int optidx = 0;
     static struct option options[] = {
       { "announce", 1, 0, 'a' },
+      { "filelist", 1, 0, 'f' },
+      { "name", 1, 0, 'n' },
       { "announcelist", 1, 0, 'A' },
+      { "webseeds", 1, 0, 'w' },
       { "piecelength", 1, 0, 'l' },
+      { "piecesize", 1, 0, 'L' },
       { "comment", 1, 0, 'c' },
       { "private", 1, 0, 'p' },
       { "nodate", 0, 0, 'D' },
@@ -1169,7 +1453,7 @@ int main(int argc, char **argv) {
       { "help", 0, 0, 'h' },
       { 0, 0, 0, 0 }
     };
-    char c = getopt_long(argc, argv, "hVqSsmCDa:A:l:c:p:", options, &optidx );
+    char c = getopt_long(argc, argv, "hVqSsmCDa:f:n:A:w:l:L:c:p:", options, &optidx );
     if (c == -1) {
       break;
     }
@@ -1179,12 +1463,24 @@ int main(int argc, char **argv) {
     case ('a'):
       url = optarg;
       break;
+    case ('f'):
+      filelistfilename = optarg;
+      break;
+    case ('n'):
+      nameflag = optarg;
+      break;
     case ('A'):
       urls = optarg;
       break;
+    case ('w'):
+      wurls = optarg;
+      break;
     case ('l'):
       plen = atoi(optarg);
       break;
+    case ('L'):
+      lplen = atoi(optarg);
+      break;
     case ('c'):
       commentstr = optarg;
       break;
@@ -1212,9 +1508,9 @@ int main(int argc, char **argv) {
       break;
     case ('V'):
       printf(
-        "buildtorrent %s (GPL) 2007-8 "
-        "Claude Heiland-Allen <claudiusmaximus@goto10.org>\n",
-        __version__
+        "buildtorrent " bt_version "\n"
+        "Copyright (C) 2007-2010 Claude Heiland-Allen <claudiusmaximus@goto10.org>\n"
+        "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
       );
       return 0;
     case ('h'):
@@ -1226,42 +1522,70 @@ int main(int argc, char **argv) {
     fprintf(stderr, "buildtorrent: announce url required\n");
     return 1;
   }
+  if (0 <= lplen && lplen < 31) {
+    plen = 1 << lplen;
+  }
   if (plen <= 0) { /* avoid division by zero */
     fprintf(stderr, "buildtorrent: piece length must be greater than 0\n");
     return 1;
   }
-  if (optind + 2 < argc) {
-    fprintf(stderr, "buildtorrent: too many arguments\n");
-    return 1;
-  }
-  if (optind + 2 > argc) {
-    fprintf(stderr, "buildtorrent: too few arguments\n");
-    return 1;
-  }
 
-  inname  = argv[optind];
-  outfile = argv[optind + 1];
+  if (filelistfilename) {
+    if (optind + 1 < argc) {
+      fprintf(stderr, "buildtorrent: too many arguments\n");
+      return 1;
+    }
+    if (optind + 1 > argc) {
+      fprintf(stderr, "buildtorrent: too few arguments\n");
+      return 1;
+    }
+    if (!nameflag) {
+      fprintf(stderr, "buildtorrent: missing '-n', required when using '-f'\n");
+      return 1;
+    }
+    inname = NULL;
+    outfile = argv[optind];
+  } else {
+    if (optind + 2 < argc) {
+      fprintf(stderr, "buildtorrent: too many arguments\n");
+      return 1;
+    }
+    if (optind + 2 > argc) {
+      fprintf(stderr, "buildtorrent: too few arguments\n");
+      return 1;
+    }
+    inname  = argv[optind];
+    outfile = argv[optind + 1];
+  }
 
   /* handle paths correctly (note: requires POSIX basename(), not GNU) */
-  strncpy(nametemp, inname, 8191);
-  nametemp[8191] = '\0';
-  namebase = basename(nametemp);
-  slen = strlen(namebase);
-  for (i = 0; i < slen; ++i) {
-    if (namebase[i] == '/') {
-      fprintf(
-        stderr,
-        "buildtorrent: BUG! input (\"%s\") munged (\"%s\") contains '/'.\n",
-        inname,
-        namebase
-      );
+  if (inname) {
+    if (strlen(inname) > 8190) {
+      fprintf(stderr, "buildtorrent: 'input' argument too long\n");
       return 1;
     }
+    strncpy(nametemp, inname, 8191);
+    nametemp[8191] = '\0';
+    namebase = basename(nametemp);
+    slen = strlen(namebase);
+    for (i = 0; i < slen; ++i) {
+      if (namebase[i] == '/') {
+        fprintf(
+          stderr,
+          "buildtorrent: BUG! input (\"%s\") munged (\"%s\") contains '/'.\n",
+          inname,
+          namebase
+        );
+        return 1;
+      }
+    }
   }
 
-  if (stat(inname, &s)) {
-    fprintf(stderr, "buildtorrent: could not stat \"%s\"\n", inname);
-    return 1;
+  if (inname) {
+    if (stat(inname, &s)) {
+      fprintf(stderr, "buildtorrent: could not stat \"%s\"\n", inname);
+      return 1;
+    }
   }
 
   if (!(torrent = bt_dictionary())) {
@@ -1280,7 +1604,12 @@ int main(int argc, char **argv) {
     fprintf(stderr, "buildtorrent: couldn't allocate piece length integer\n");
     return 1;
   }
-  if (!(name = bt_string(strlen(namebase), namebase))) {
+  if (nameflag) {
+    name = bt_string(strlen(nameflag), nameflag);
+  } else {
+    name = bt_string(strlen(namebase), namebase);
+  }
+  if (!name) {
     fprintf(stderr, "buildtorrent: couldn't allocate name string\n");
     return 1;
   }
@@ -1300,83 +1629,82 @@ int main(int argc, char **argv) {
       return 1;
     }
   }
-
-  if (!(queue = bt_create_hash_queue())) {
-    fprintf(stderr, "buildtorrent: couldn't allocate hash queue\n");
+  if (wurls) {
+    if (!(webseedlist = bt_parse_webseedlist(wurls))) {
+      fprintf(stderr, "buildtorrent: error parsing webseed list argument\n");
+      return 1;
+    }
+  }
+  if (!(flist = bt_create_file_list())) {
+    fprintf(stderr, "buildtorrent: couldn't allocate file list\n");
+    return 1;
+  }
+  if (!(aflist = bt_create_afile_list())) {
+    fprintf(stderr, "buildtorrent: couldn't allocate annotated file list\n");
     return 1;
   }
 
-  if (S_ISDIR(s.st_mode)) {
+  if (inname && S_ISDIR(s.st_mode)) {
 
+    multifile = 1;
     if (!(dir = opendir(inname))) {
       fprintf(stderr, "buildtorrent: couldn't open directory\n");
       return 1;
     }
-    if (!(files = bt_list())) {
-      fprintf(stderr, "buildtorrent: couldn't allocate files list\n");
-      return 1;
-    }
     if (!(pathlist = bt_list())) {
       fprintf(stderr, "buildtorrent: couldn't path list\n");
       return 1;
     }
-    if (strlen(inname) > 8190) {
-      fprintf(stderr, "buildtorrent: 'input' argument too long\n");
-      return 1;
-    }
     memcpy(path, inname, strlen(inname) + 1);
-    if (bt_findfiles(
-      dir, 8192, path, pathlist, files, queue, domd5sum, verbose
-    )) {
+    if (bt_find_files(dir, 8192, path, pathlist, flist)) {
       fprintf(stderr, "buildtorrent: error finding files\n");
       return 1;
     }
     closedir(dir);
-    if (bt_dictionary_insert(info, strlen("files"), "files", files)) {
-      fprintf(stderr, "buildtorrent: couldn't insert files into info\n");
+
+  } else if (inname && S_ISREG(s.st_mode)) {
+
+    multifile = 0;
+    if ((bt_file_list_prepend(flist, inname, NULL))) {
+      fprintf(stderr, "buildtorrent: error building single file list\n");
       return 1;
     }
 
-  } else if (S_ISREG(s.st_mode)) {
-
-    if (!(length = bt_integer(s.st_size))) {
-      fprintf(stderr, "buildtorrent: couldn't allocate length integer\n");
-      return 1;
-    }
-    if (bt_dictionary_insert(info, strlen("length"), "length", length)) {
-      fprintf(stderr, "buildtorrent: couldn't insert length into info\n");
-      return 1;
-    }
-    if (domd5sum) {
-      if (!(md5sum = bt_md5sum_file(inname, s.st_size))) {
-        fprintf(stderr, "buildtorrent: couldn't md5sum \"%s\"\n", inname);
-        return 1;
-      }
-      if (bt_dictionary_insert(info, strlen("md5sum"), "md5sum", md5sum)) {
-        fprintf(stderr, "buildtorrent: couldn't insert md5sum into info\n");
-        return 1;
+  } else if (!inname) {
+ 
+    multifile = 1;
+    bt_error err = BT_ERROR_NONE;
+    if (0 == strcmp("-", filelistfilename)) {
+      err = bt_parse_filelist(flist, stdin);
+    } else {
+      FILE *filelistfile;
+      if ((filelistfile = fopen(filelistfilename, "rb"))) {
+        err = bt_parse_filelist(flist, filelistfile);
+        fclose(filelistfile);
+      } else {
+        fprintf(stderr, "buildtorrent: couldn't open file list \"%s\"\n", filelistfilename);
+        return 1;        
       }
     }
-    if ((bt_hash_enqueue(queue, inname, s.st_size, verbose))) {
-      fprintf(stderr, "buildtorrent: error enqueueing file for hashing\n");
+    if (err) {
+      fprintf(stderr, "buildtorrent: error processing file list\n");
       return 1;
     }
 
   } else {
+
     fprintf(
       stderr, "buildtorrent: \"%s\" is neither file nor directory\n", inname
     );
     return 1;
+
   }
 
-  if (!(pieces = bt_hash_pieces(queue, plen, verbose))) {
-    fprintf(stderr, "buildtorrent: error hashing files\n");
-    return 1;
-  }
-  if (bt_dictionary_insert(info, strlen("pieces"), "pieces", pieces)) {
-    fprintf(stderr, "buildtorrent: couldn't insert pieces into info\n");
+  if ((bt_annotate_files(flist, aflist))) {
+    fprintf(stderr, "buildtorrent: error annotating file list\n");
     return 1;
   }
+
   if (privated) {
     if (!(private = bt_integer(privateopt))) {
       fprintf(stderr, "buildtorrent: couldn't allocate private integer\n");
@@ -1397,9 +1725,11 @@ int main(int argc, char **argv) {
       return 1;
     }
   }
-  if (bt_dictionary_insert(torrent, strlen("info"), "info", info)) {
-    fprintf(stderr, "buildtorrent: couldn't insert info into torrent\n");
-    return 1;
+  if (wurls) {
+    if (bt_dictionary_insert(torrent, strlen("url-list"), "url-list", webseedlist)) {
+      fprintf(stderr, "buildtorrent: couldn't insert webseed url-list into torrent\n");
+      return 1;
+    }
   }
   if (!nodate) {
     if (!(creationdate = bt_integer(time(NULL)))) {
@@ -1414,7 +1744,7 @@ int main(int argc, char **argv) {
     }
   }
   if (!nocreator) {
-    if (!(creator = bt_string(strlen("buildtorrent/0.7"), "buildtorrent/0.7"))) {
+    if (!(creator = bt_string(strlen("buildtorrent/" bt_version), "buildtorrent/" bt_version))) {
       fprintf(stderr, "buildtorrent: couldn't allocate created by string\n");
       return 1;
     }
@@ -1441,6 +1771,48 @@ int main(int argc, char **argv) {
     fprintf(stderr, "buildtorrent: couldn't open \"%s\" for writing\n", outfile);
     return 1;
   }
+
+  if (!(pieces = bt_hash_pieces(aflist, plen, domd5sum, verbose))) {
+    fprintf(stderr, "buildtorrent: error hashing files\n");
+    return 1;
+  }
+
+  if (multifile) {
+    if (!(files = bt_afile_list_info(aflist))) {
+      fprintf(stderr, "buildtorrent: error getting file list info\n");
+      return 1;
+    }
+    if (bt_dictionary_insert(info, strlen("files"), "files", files)) {
+      fprintf(stderr, "buildtorrent: couldn't insert files into info\n");
+      return 1;
+    }
+  } else {
+    bt_afile_list node = aflist->next;
+    if (!(length = bt_integer(node->length))) {
+      fprintf(stderr, "buildtorrent: couldn't allocate length integer\n");
+      return 1;
+    }
+    if (bt_dictionary_insert(info, strlen("length"), "length", length)) {
+      fprintf(stderr, "buildtorrent: couldn't insert length into info\n");
+      return 1;
+    }
+    if (node->md5sum) {
+      if (bt_dictionary_insert(info, strlen("md5sum"), "md5sum", node->md5sum)) {
+        fprintf(stderr, "buildtorrent: couldn't insert md5sum into info\n");
+        return 1;
+      }
+    }
+  }
+
+  if (bt_dictionary_insert(info, strlen("pieces"), "pieces", pieces)) {
+    fprintf(stderr, "buildtorrent: couldn't insert pieces into info\n");
+    return 1;
+  }
+  if (bt_dictionary_insert(torrent, strlen("info"), "info", info)) {
+    fprintf(stderr, "buildtorrent: couldn't insert info into torrent\n");
+    return 1;
+  }
+
   if (bt_write(output, torrent)) {
     fprintf(stderr, "buildtorrent: error writing \"%s\"\n", outfile);
     return 1;
@@ -1448,7 +1820,6 @@ int main(int argc, char **argv) {
   if (show) {
     printf("torrent =>\n");
     bt_show(torrent, show == 2, 2, 2, 0);
-    printf("}\n");
   }
   bt_free(torrent);
   fclose(output);
